package servertest;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javaslang.Tuple;

import org.apache.commons.io.IOUtils;
import org.testng.Assert;

public class Server {
	
	private static final String STDERR_TXT = "stderr.txt";
	private static final String STDOUT_TXT = "stdout.txt";
	protected Path sourceFolder;
	protected Path serverHomeFolder;
	
	protected Path pomXml;
	protected Path targetFolder;
	protected Path targetJar;
	
	protected Process serverProcess;
	protected InputStream serverStdOutput;
	protected InputStream serverErrorOutput;
	
	protected Runtime runtime;
	
	public Server(Path sourceFolder, Path serverHomeFolder) throws IOException, InterruptedException {
		this.serverHomeFolder = serverHomeFolder;
		this.sourceFolder = sourceFolder;
		this.pomXml = sourceFolder.resolve(Paths.get("pom.xml"));
		
		if(!pomXml.toFile().exists()) {
			List<File> dirs = Stream.of(sourceFolder.toFile().listFiles())
			.filter(File::isDirectory)
			.collect(Collectors.toList());
			
			System.out.printf("Dirs: %s\n", dirs);
			
			if(dirs.size() == 1) {
				this.sourceFolder = sourceFolder.resolve(dirs.get(0).toPath());
				pomXml = this.sourceFolder.resolve("pom.xml");
			}
		}

		FileAssertUtil.assertExists(pomXml);
		this.targetFolder = this.sourceFolder.resolve(Paths.get("target"));
		
		if(!targetFolder.toFile().exists()) {
			// compile if not exists
			Compiler.compile(pomXml);
		}

		FileAssertUtil.assertExists(this.sourceFolder);
		FileAssertUtil.assertExists(this.serverHomeFolder);
		//FileAssertUtil.assertExists(this.sourceFolder.resolve(Paths.get(".htaccess")));
	}

	public Path getServerSourceHome() {
		return sourceFolder;
	}
	
	public void run(int port) throws IOException, InterruptedException {
		try {
			targetJar = getJarOfServer();
		}catch(NoSuchElementException nse) {
			System.err.printf("Server %s has no jars!\n", this.targetFolder);
			return;
		}
		
		String runCommand = String.format("java -jar %s %d %s", targetJar.toAbsolutePath().toString(), port, serverHomeFolder.toAbsolutePath().toString());
		//String runCommand = "python runner.py";
		System.out.printf("Running server with command: '%s'\n", runCommand);
		
		// Run a java app in a separate system process
		runtime = Runtime.getRuntime();
		serverProcess = runtime.exec(runCommand);
		serverProcess.getOutputStream().close();
		
		// wait for server to start
		Thread.sleep(2000);
		
		if(serverProcess.isAlive()) {
			System.out.println("Server running.");
		}else {
			writeServerOutput();
			throw new RuntimeException("Server died! See '"+STDOUT_TXT+"' and '"+STDERR_TXT+"' files.");
		}
	}
	
	public void stop() throws FileNotFoundException, IOException {
		System.out.println("Writing output of server process: " + this.sourceFolder);
		//writeServerOutput();
		System.out.println("Stopping server process: " + this.sourceFolder);
		if(serverProcess != null) {
			serverProcess.destroyForcibly();
		}else {
			System.err.println("Server was null: " + this.sourceFolder);
		}
	}
	
	private void writeServerOutput() throws FileNotFoundException, IOException {
		InputStream inputStream = serverProcess.getInputStream();
		IOUtils.copy(inputStream, new FileOutputStream(STDOUT_TXT));
		inputStream.close();
		
		InputStream errorStream = serverProcess.getErrorStream();
		IOUtils.copy(errorStream, new FileOutputStream(STDERR_TXT));
		errorStream.close();
	}

	private Path getJarOfServer() throws IOException {
		List<Path> jars = findJars(".jar");

		// take the biggest jar as the correct one
		return jars.stream()
		.map(path -> Tuple.of(path, path.toFile().length()))
		.reduce((p1, p2) -> p1._2 > p2._2 ? p1 : p2)
		.map(p -> p._1)
		.get();
	}

	private List<Path> findJars(String suffix) throws IOException {
		List<Path> jars;
		jars = Files
		.list(targetFolder)
		.filter(path -> path.toString().endsWith(suffix))
		.map(a ->{ System.out.println(a); return a;})
		.collect(Collectors.toList());
		return jars;
	}
	
}
