package servertest;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Base64;
import java.util.stream.Stream;



import javaslang.Tuple;
import javaslang.Tuple2;



import org.apache.commons.io.FileUtils;
import org.jsoup.Connection;
import org.jsoup.Connection.Response;
import org.jsoup.Jsoup;
import org.testng.Assert;

public class Test {
	
	private static final String REPORT_FILENAME = "functional-report.html";

	/**
	 * Timeout in millis.
	 */
	private static final int REQUEST_TIMEOUT = 1000;
	
	protected TestReport testReport;
	protected Server server;
	protected int port;
	protected Path testWebRoot = Paths.get("home");
	
	public Test(int port) throws IOException {
		this.port = port;
		
		// run this when server is already running at given port
		TestReport report = new TestReport();
		try {
			tests(report);
		} catch(Throwable t) {
			t.printStackTrace();
			report.addTestResult("Check that server can be started.", () -> { throw t; });
		} finally {
			System.out.println("All tests done. Exiting");
			report.save(REPORT_FILENAME);
		}
	}
	
	public Test(String serverSourceHome, int port) throws FileNotFoundException, IOException {
		this.port = port;
		
		TestReport report = new TestReport();
		server = null;
		try {
			server = new Server(Paths.get(serverSourceHome), testWebRoot);
			server.run(port);
			
			if(!prepareHtaccess()) {
				System.err.println("No htaccess file in root!");
				report.addError("Htaccess with credentials test:123456 exists.", "FAIL");
			}
			
			tests(report);
		} catch(Throwable t) {
			t.printStackTrace();
			report.addTestResult("Check that server can be started.", () -> { throw t; });
		} finally {
			System.out.println("All tests done. Exiting");
			if(server != null) {
				server.stop();
			}
			report.save(REPORT_FILENAME);
		}
	}
	 
	private boolean prepareHtaccess() throws IOException {
		File providedHtaccess = server.getServerSourceHome().resolve(".htaccess").toFile();
		
		if(providedHtaccess.exists()) {
		// copy htaccess with username "test" and password "123456"
			FileUtils.copyFile(server.getServerSourceHome().resolve(".htaccess").toFile(), testWebRoot.resolve("secureFolder").resolve(".htaccess").toFile());
			return true;
		}else {
			return false;
		}
	}

	public void tests(TestReport rep) throws IOException {
		pageHasElementWithText(rep, "http://localhost:"+port+"/index.html", "h1", "Hello world!");
		pageHasElementWithText(rep, "http://localhost:"+port+"/", "h1", "Hello world!");
		pageHasElementWithText(rep, "http://localhost:"+port+"/folder/pageInFolder.html", "b", "Page in folder");
		pageHasElementWithText(rep, "http://localhost:"+port+"/file with spaces.html", "strong", "File with spaces in name");
		
		// check query string
		hasStatus(rep, "http://localhost:"+port+"/index.html?param", 200);
		hasStatus(rep, "http://localhost:"+port+"/index.html?param=value", 200);
		hasStatus(rep, "http://localhost:"+port+"/index.html#heading1", 200);
		hasStatus(rep, "http://localhost:"+port+"/index.html?param=value#heading1", 200);
		
		// check nonexisting file
		hasStatus(rep, "http://localhost:"+port+"/nonExisting.html", 404);
		
		// check unreachable files
		hasStatus(rep, "http://localhost:"+port+"/../../pom.xml", 404);
		//hasStatus(rep, "http://localhost:"+port+"/C:\\windows\\system32\\", 404);
		
		// check security - unauthorized
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secure.html", 401);
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secureSubFolder/inSecureSubFolder.html", 401);
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/.htaccess", 404);
		hasHeader(rep, "http://localhost:"+port+"/secureFolder/secure.html", "WWW-Authenticate");
		
		// check security - wrong authorization
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secure.html", 401, "wrongUsername", "wrongPassword");
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secure.html", 401, "test", "wrongPassword");
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secure.html", 401, "wrongUsername", "123456");
		
		// check security - correct authorization
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secure.html", 200, "test", "123456");
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/.htaccess", 404, "test", "123456");
		hasStatus(rep, "http://localhost:"+port+"/secureFolder/secureSubFolder/inSecureSubFolder.html", 200, "test", "123456");
		
		// check image download
		downloadAndCheckFile(rep, "http://localhost:"+port+"/image.jpg", "home/image.jpg");
		
		// check long requests
		Stream.of(129, 1025, 2000, 8000).forEach(length -> {			
			hasStatus(rep, createLongUrl("http://localhost:"+port+"/index.html?long=", length), 200, null, null, "Trying request 'index.html?long=aaa...' that is "+length+" characters long.");
		});
	}
		
    public static void pageHasElementWithText(TestReport rep, String page, String elementSelector, String text) throws IOException {
    	pageHasElementWithText(rep, page, elementSelector, text, null, null);
    }
    public static void pageHasElementWithText(TestReport rep, String page, String elementSelector, String text, String username, String password) throws IOException {
    	hasStatus(rep, page, 200, username, password);
    	String desc = String.format("Check that page '%s' has element with css selector '%s' with text: '%s'.", page, elementSelector, text);
    	System.out.printf("Running test: %s\n", desc);
    	rep.addTestResult(desc, () -> Assert.assertEquals(Jsoup.connect(page).get().select(elementSelector).text(), text));
    }
    
    public static void hasStatus(TestReport rep, String page, int expectedStatusCode) {
    	hasStatus(rep, page, expectedStatusCode, null, null);
    }
    public static void hasStatus(TestReport rep, String page, int expectedStatusCode, String username, String password) {
    	String desc = String.format("Check that page '%s' returns status code '%d' with authorization: '%s'.", page, expectedStatusCode, username+":"+password);
    	hasStatus(rep, page, expectedStatusCode, username, password, desc);
    }
    public static void hasStatus(TestReport rep, String page, int expectedStatusCode, String username, String password, String description) {
    	System.out.printf("Running test: %s\n", description);
    	Tuple2<String, Connection> authAndConn = makeConnection(page, username, password);
    	rep.addTestResult(description, () -> {
    		// make each status check twice to see whether server have crashed or not
    		Assert.assertEquals(authAndConn._2.ignoreHttpErrors(true).execute().statusCode(), expectedStatusCode);
    		Assert.assertEquals(authAndConn._2.ignoreHttpErrors(true).execute().statusCode(), expectedStatusCode);
    	});
    }
    
    public static void hasHeader(TestReport rep, String page, String expectedHeader) {
    	Tuple2<String, Connection> authAndConn = makeConnection(page, null, null);
    	String desc = String.format("Check that page '%s' returns response header '%s' with authorization: '%s'.", page, expectedHeader, authAndConn._1);
    	System.out.printf("Running test: %s\n", desc);
    	rep.addTestResult(desc, () -> {
    		Response res = authAndConn._2.ignoreHttpErrors(true).execute();
    		System.out.println(res.headers().keySet());
    		Assert.assertTrue(res.headers().containsKey(expectedHeader));
    	});
    }

	private static Tuple2<String, Connection> makeConnection(String page, String username, String password) {
		Connection jsoupConnection = Jsoup.connect(page).timeout(REQUEST_TIMEOUT);
		String authorization = "none";
		if(username != null && password != null) {
    		authorization = username + ":" + password;
    		String base64login = new String(Base64.getEncoder().encodeToString(authorization.getBytes()));
    		jsoupConnection = jsoupConnection.header("Authorization", "Basic " + base64login);
    	}
		return Tuple.of(authorization, jsoupConnection);
	}
    
	private static String createLongUrl(String prefix, int targetLength) {
		StringBuilder sb = new StringBuilder(prefix);
		while(sb.length() < targetLength) {
			sb.append("a");
		}
		return sb.toString();
	}
	
    private static void downloadAndCheckFile(TestReport rep, String fileUrl, String filePath) {
    	String desc = String.format("Download binary file '%s' and check that it is the same as '%s'.", fileUrl, filePath);
    	rep.addTestResult(desc, () -> {
    		File referenceFile = new File(filePath);
    		Assert.assertTrue(referenceFile.exists(), "Reference file does not exist (this is not fault of server, but bug in the test)!");
    		
    		File tempFile = new File("tempFile");
    		if(tempFile.exists()) {
    			tempFile.delete();
    		}
    		Assert.assertTrue(!tempFile.exists(), "Temp file already exists (this is not fault of server, but bug in the test)!");
    		
    		FileUtils.copyURLToFile(new URL(fileUrl), tempFile, REQUEST_TIMEOUT, REQUEST_TIMEOUT);
    		Assert.assertTrue(FileUtils.contentEquals(referenceFile, tempFile), "Downloaded file and reference file are not the same!");
    	});
    }

    
    
}
