package servertest;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import lombok.Value;

@Value
class TestRun {
	protected String description;
	protected String result;
}

interface TestFunction {
	public void apply() throws Exception;
}

public class TestReport {
	
	protected List<TestRun> runs = new LinkedList<>();
	protected int errorCount = 0;
	
	public void addTestResult(String description, TestFunction fun) {
		try {
			fun.apply();
			
			// no exception thrown, result was OK
			runs.add(new TestRun(description, "OK"));
		} catch (Throwable e) {
			errorCount++;
			String exceptionDescription = String.format("Exception %s with message '%s'", e.getClass().getName(), e.getMessage());
			runs.add(new TestRun(description, exceptionDescription));
		}
	}
	
	public void addError(String description, String result) {
		errorCount++;
		runs.add(new TestRun(description, result));
	}
	
	public void save(String filename) {
		try(ReportWriter writer = new ReportWriter(filename)) {
			writer.println("<h2>Functional tests</h2>");
			writer.printf("<b>%d / %d tests failed</b><br>", errorCount, runs.size());
			writer.encloseTag(
					"table", 
					"border=1",
					runs.stream()
					.map(run -> String.format("<tr><td>%s</td><td>%s</td></tr>", handleTextOverflow(run.getDescription(), 150), run.getResult()))
					.collect(Collectors.joining("\n"))
			);
			
			int points = Math.round((1 - ((float)errorCount / runs.size())) * 10);
			writer.printf("<br><b>Hodnocen: </b> %d / %d <br>\n", points, 10);
		}catch(IOException ioe) {
			throw new UncheckedIOException(ioe);
		}
	}
	
	private static String handleTextOverflow(String str, int limit) {
		/*if(str.length() > limit) {
			return str.substring(0,  limit) + "...";
		}*/
		return str;
	}
	
}
